@@ -1,11 +1,12 @@
 pub use anyhow::{anyhow, Context as _};
 pub use tracing::{trace, debug, info, warn, error, Level};
 pub use tracing_tools::{span, TracingTask, PinnedFut};
-pub use async_channel::{Sender, Receiver, RecvError, bounded as bounded_ch, unbounded as unbounded_ch};
+pub use crusty_core::flume::{Sender, Receiver, RecvError, bounded as bounded_ch, unbounded as unbounded_ch};
 pub use tokio::time::{self, Instant, Duration};
 pub use url::Url;
 
 pub use std::{
+    rc::Rc,
     sync::{Arc, Mutex},
     collections::{HashMap, HashSet, LinkedList},
     iter::Iterator,
